#!/usr/bin/env python

"""
<Program Name>
  pycrypto_keys.py

<Author>
  Vladimir Diaz <vladimir.v.diaz@gmail.com>

<Started>
  October 7, 2013.

<Copyright>
  See LICENSE for licensing information.

<Purpose>
  The goal of this module is to support public-key and general-purpose
  cryptography through the PyCrypto library.  The RSA-related functions provided:
  generate_rsa_public_and_private()
  create_rsa_signature()
  verify_rsa_signature()
  create_rsa_encrypted_pem()
  create_rsa_public_and_private_from_pem()

  The general-purpose functions include:
  encrypt_key()
  decrypt_key()

  PyCrypto (i.e., the 'Crypto' package) performs the actual cryptographic
  operations and the functions listed above can be viewed as the easy-to-use
  public interface.

  https://github.com/dlitz/pycrypto
  https://en.wikipedia.org/wiki/RSA_(algorithm)
  https://en.wikipedia.org/wiki/Advanced_Encryption_Standard
  https://en.wikipedia.org/wiki/3des
  https://en.wikipedia.org/wiki/PBKDF

  TUF key files are encrypted with the AES-256-CTR-Mode symmetric key
  algorithm.  User passwords are strengthened with PBKDF2, currently set to
  100,000 passphrase iterations.  The previous evpy implementation used 1,000
  iterations.

  PEM-encrypted RSA key files use the Triple Data Encryption Algorithm (3DES)
  and Cipher-block chaining (CBC) for the mode of operation.  Password-Based Key
  Derivation Function 1 (PBKF1) + MD5.
 """

# Help with Python 3 compatibility, where the print statement is a function, an
# implicit relative import is invalid, and the '/' operator performs true
# division.  Example:  print 'hello world' raises a 'SyntaxError' exception.
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

import os
import binascii
import json

# Crypto.PublicKey (i.e., PyCrypto's public-key cryptography modules) supports
# algorithms like the Digital Signature Algorithm (DSA) and the ElGamal
# encryption system.  'Crypto.PublicKey.RSA' is needed here to generate, sign,
# and verify RSA keys.
import Crypto.PublicKey.RSA

# PyCrypto requires 'Crypto.Hash' hash objects to generate PKCS#1 PSS
# signatures (i.e., Crypto.Signature.PKCS1_PSS).
import Crypto.Hash.SHA256

# RSA's probabilistic signature scheme with appendix (RSASSA-PSS).
# PKCS#1 v1.5 is available for compatibility with existing applications, but
# RSASSA-PSS is encouraged for newer applications.  RSASSA-PSS generates
# a random salt to ensure the signature generated is probabilistic rather than
# deterministic (e.g., PKCS#1 v1.5).
# http://en.wikipedia.org/wiki/RSA-PSS#Schemes
# https://tools.ietf.org/html/rfc3447#section-8.1
import Crypto.Signature.PKCS1_PSS

# Import PyCrypto's Key Derivation Function (KDF) module.  'keys.py' needs this
# module to derive a secret key according to the Password-Based Key Derivation
# Function 2 specification.  The derived key is used as the symmetric key to
# encrypt TUF key information.  PyCrypto's implementation:
# Crypto.Protocol.KDF.PBKDF2().  PKCS#5 v2.0 PBKDF2 specification:
# http://tools.ietf.org/html/rfc2898#section-5.2
import Crypto.Protocol.KDF

# PyCrypto's AES implementation.  AES is a symmetric key algorithm that
# operates on fixed block sizes of 128-bits.
# https://en.wikipedia.org/wiki/Advanced_Encryption_Standard
import Crypto.Cipher.AES

# 'Crypto.Random' is a cryptographically strong version of Python's standard
# "random" module.  Random bits of data is needed for salts and
# initialization vectors suitable for the encryption algorithms used in
# 'pycrypto_keys.py'.
import Crypto.Random

# The mode of operation is presently set to CTR (CounTeR Mode) for symmetric
# block encryption (AES-256, where the symmetric key is 256 bits).  PyCrypto
# provides a callable stateful block counter that can update successive blocks
# when needed.  The initial random block, or initialization vector (IV), can
# be set to begin the process of incrementing the 128-bit blocks and allowing
# the AES algorithm to perform cipher block operations on them.
import Crypto.Util.Counter

import securesystemslib.exceptions
import securesystemslib.hash
import securesystemslib.formats
import securesystemslib.util

# Extract the cryptography library settings.
import securesystemslib.settings

# Recommended RSA key sizes:
# http://www.emc.com/emc-plus/rsa-labs/historical/twirl-and-rsa-key-size.htm#table1
# According to the document above, revised May 6, 2003, RSA keys of
# size 3072 provide security through 2031 and beyond.
_DEFAULT_RSA_KEY_BITS = 3072

# The delimiter symbol used to separate the different sections
# of encrypted files (i.e., salt, iterations, hmac, IV, ciphertext).
# This delimiter is arbitrarily chosen and should not occur in
# the hexadecimal representations of the fields it is separating.
_ENCRYPTION_DELIMITER = '@@@@'

# AES key size.  Default key size = 32 bytes = AES-256.
_AES_KEY_SIZE = 32

# Default salt size, in bytes.  A 128-bit salt (i.e., a random sequence of data
# to protect against attacks that use precomputed rainbow tables to crack
# password hashes) is generated for PBKDF2.
_SALT_SIZE = 16

# Default PBKDF2 passphrase iterations.  The current "good enough" number
# of passphrase iterations.  We recommend that important keys, such as root,
# be kept offline.  'settings.PBKDF2_ITERATIONS' should increase as CPU
# speeds increase, set here at 100,000 iterations by default (in 2013).
# Repository maintainers may opt to modify the default setting according to
# their security needs and computational restrictions.  A strong user password
# is still important.  Modifying the number of iterations will result in a new
# derived key+PBDKF2 combination if the key is loaded and re-saved, overriding
# any previous iteration setting used by the old '<keyid>.key'.
# https://en.wikipedia.org/wiki/PBKDF2
_PBKDF2_ITERATIONS = securesystemslib.settings.PBKDF2_ITERATIONS


def generate_rsa_public_and_private(bits=_DEFAULT_RSA_KEY_BITS):
  """
  <Purpose>
    Generate public and private RSA keys with modulus length 'bits'.  The
    public and private keys returned conform to
    'securesystemslib.formats.PEMRSA_SCHEMA' and have the form:

    '-----BEGIN RSA PUBLIC KEY----- ...'

    or

    '-----BEGIN RSA PRIVATE KEY----- ...'

    The public and private keys are returned as strings in PEM format.

    Although PyCrypto sets a 1024-bit minimum key size,
    generate_rsa_public_and_private() enforces a minimum key size of 2048 bits.
    If 'bits' is unspecified, a 3072-bit RSA key is generated, which is the key
    size recommended by TUF.

    >>> public, private = generate_rsa_public_and_private(2048)
    >>> securesystemslib.formats.PEMRSA_SCHEMA.matches(public)
    True
    >>> securesystemslib.formats.PEMRSA_SCHEMA.matches(private)
    True

  <Arguments>
    bits:
      The key size, or key length, of the RSA key.  'bits' must be 2048, or
      greater, and a multiple of 256.

  <Exceptions>
    securesystemslib.exceptions.FormatError, if 'bits' does not contain the correct
    format.

    ValueError, if an exception occurs in the RSA key generation routine.
    'bits' must be a multiple of 256.  The 'ValueError' exception is raised by
    the PyCrypto key generation function.

  <Side Effects>
    The RSA keys are generated by PyCrypto's Crypto.PublicKey.RSA.generate().

  <Returns>
    A (public, private) tuple containing the RSA keys in PEM format.
  """

  # Does 'bits' have the correct format?
  # This check will ensure 'bits' conforms to
  # 'securesystemslib.formats.RSAKEYBITS_SCHEMA'.  'bits' must be an integer object,
  # with a minimum value of 2048.  Raise 'securesystemslib.exceptions.FormatError'
  # if the check fails.
  securesystemslib.formats.RSAKEYBITS_SCHEMA.check_match(bits)

  # Generate the public and private RSA keys.  The PyCrypto module performs
  # the actual key generation.  Raise 'ValueError' if 'bits' is less than 1024
  # or not a multiple of 256, although a 2048-bit minimum is enforced by
  # securesystemslib.formats.RSAKEYBITS_SCHEMA.check_match().
  rsa_key_object = Crypto.PublicKey.RSA.generate(bits)

  # Extract the public & private halves of the RSA key and generate their
  # PEM-formatted representations.  Return the key pair as a (public, private)
  # tuple, where each RSA is a string in PEM format.
  private = rsa_key_object.exportKey(format='PEM')
  rsa_pubkey = rsa_key_object.publickey()
  public = rsa_pubkey.exportKey(format='PEM')

  return public.decode(), private.decode()





def create_rsa_signature(private_key, data):
  """
  <Purpose>
    Generate an RSASSA-PSS signature.  The signature, and the method (signature
    algorithm) used, is returned as a (signature, method) tuple.

    The signing process will use 'private_key' and 'data' to generate the
    signature.

    RFC3447 - RSASSA-PSS
    http://www.ietf.org/rfc/rfc3447.txt

    >>> public, private = generate_rsa_public_and_private(2048)
    >>> data = 'The quick brown fox jumps over the lazy dog'.encode('utf-8')
    >>> signature, method = create_rsa_signature(private, data)
    >>> securesystemslib.formats.NAME_SCHEMA.matches(method)
    True
    >>> method == 'RSASSA-PSS'
    True
    >>> securesystemslib.formats.PYCRYPTOSIGNATURE_SCHEMA.matches(signature)
    True

  <Arguments>
    private_key:
      The private RSA key, a string in PEM format.

    data:
      Data (string) used by create_rsa_signature() to generate the signature.

  <Exceptions>
    securesystemslib.exceptions.FormatError, if 'private_key' is improperly formatted.

    TypeError, if 'private_key' is unset.

    securesystemslib.exceptions.CryptoError, if the signature cannot be generated.

  <Side Effects>
    PyCrypto's 'Crypto.Signature.PKCS1_PSS' called to generate the signature.

  <Returns>
    A (signature, method) tuple, where the signature is a string and the method
    is 'RSASSA-PSS'.
  """

  # Does 'private_key' have the correct format?
  # This check will ensure 'private_key' conforms to
  # 'securesystemslib.formats.PEMRSA_SCHEMA'.  Raise
  # 'securesystemslib.exceptions.FormatError' if the check fails.
  securesystemslib.formats.PEMRSA_SCHEMA.check_match(private_key)

  # Does 'data' have the correct format?
  securesystemslib.formats.DATA_SCHEMA.check_match(data)

  # Signing the 'data' object requires a private key.
  # The 'RSASSA-PSS' (i.e., PyCrypto module) signing method is the
  # only method currently supported.
  method = 'RSASSA-PSS'
  signature = None

  # Verify the signature, but only if the private key has been set.  The
  # private key is a NULL string if unset.  Although it may be clearer to
  # explicitly check that 'private_key' is not '', we can/should check for a
  # value and not compare identities with the 'is' keyword.  Up to this point
  # 'private_key' has variable size and can be an empty string.
  if len(private_key):
    # Calculate the SHA256 hash of 'data' and generate the hash's PKCS1-PSS
    # signature.

    # PyCrypto's expected exceptions when generating RSA key object:
    # "ValueError/IndexError/TypeError:  When the given key cannot be parsed
    # (possibly because the passphrase is wrong)."
    # If the passphrase is incorrect, PyCrypto returns: "RSA key format is not
    # supported".
    try:
      sha256_object = Crypto.Hash.SHA256.new(data)
      rsa_key_object = Crypto.PublicKey.RSA.importKey(private_key)

    except (ValueError, IndexError, TypeError) as e:
      raise securesystemslib.exceptions.CryptoError('Invalid private key or'
        ' hash data: ' + str(e))

    # Generate RSSA-PSS signature.  Raise 'securesystemslib.exceptions.CryptoError'
    # for the expected PyCrypto exceptions.
    try:
      pkcs1_pss_signer = Crypto.Signature.PKCS1_PSS.new(rsa_key_object)
      signature = pkcs1_pss_signer.sign(sha256_object)

    except ValueError: #pragma: no cover
      raise securesystemslib.exceptions.CryptoError('The RSA key too small for'
        ' given hash algorithm.')

    except TypeError:
      raise securesystemslib.exceptions.CryptoError('Missing required RSA'
        ' private key.')

    except IndexError: # pragma: no cover
      raise securesystemslib.exceptions.CryptoError('An RSA signature cannot'
        ' be generated: ' + str(e))

  else:
    raise ValueError('The required private key is unset.')

  return signature, method





def verify_rsa_signature(signature, signature_method, public_key, data):
  """
  <Purpose>
    Determine whether the corresponding private key of 'public_key' produced
    'signature'.  verify_signature() will use the public key, signature method,
    and 'data' to complete the verification.

    >>> public, private = generate_rsa_public_and_private(2048)
    >>> data = b'The quick brown fox jumps over the lazy dog'
    >>> signature, method = create_rsa_signature(private, data)
    >>> verify_rsa_signature(signature, method, public, data)
    True
    >>> verify_rsa_signature(signature, method, public, b'bad_data')
    False

  <Arguments>
    signature:
      An RSASSA PSS signature as a string.  This is the signature returned
      by create_rsa_signature().

    signature_method:
      A string that indicates the signature algorithm used to generate
      'signature'.  'RSASSA-PSS' is currently supported.

    public_key:
      The RSA public key, a string in PEM format.

    data:
      Data object used by securesystemslib.keys.create_signature() to generate
      'signature'.  'data' is needed here to verify the signature.

  <Exceptions>
    securesystemslib.exceptions.UnknownMethodError.  Raised if the signing method
    used by 'signature' is not one supported by
    securesystemslib.keys.create_signature().

    securesystemslib.exceptions.FormatError. Raised if 'signature',
    'signature_method', or 'public_key' is improperly formatted.

  <Side Effects>
    Crypto.Signature.PKCS1_PSS.verify() called to do the actual verification.

  <Returns>
    Boolean.  True if the signature is valid, False otherwise.
  """

  # Does 'public_key' have the correct format?
  # This check will ensure 'public_key' conforms to
  # 'securesystemslib.formats.PEMRSA_SCHEMA'.  Raise
  # 'securesystemslib.exceptions.FormatError' if the check fails.
  securesystemslib.formats.PEMRSA_SCHEMA.check_match(public_key)

  # Does 'signature_method' have the correct format?
  securesystemslib.formats.NAME_SCHEMA.check_match(signature_method)

  # Does 'signature' have the correct format?
  securesystemslib.formats.PYCRYPTOSIGNATURE_SCHEMA.check_match(signature)

  # Does 'data' have the correct format?
  securesystemslib.formats.DATA_SCHEMA.check_match(data)

  # Verify whether the private key of 'public_key' produced 'signature'.
  # Before returning the 'valid_signature' Boolean result, ensure 'RSASSA-PSS'
  # was used as the signing method.
  valid_signature = False

  # Verify the signature with PyCrypto if the signature method is valid,
  # otherwise raise 'securesystemslib.exceptions.UnknownMethodError'.
  if signature_method == 'RSASSA-PSS':
    try:
      rsa_key_object = Crypto.PublicKey.RSA.importKey(public_key)
      pkcs1_pss_verifier = Crypto.Signature.PKCS1_PSS.new(rsa_key_object)
      sha256_object = Crypto.Hash.SHA256.new(data)
      valid_signature = pkcs1_pss_verifier.verify(sha256_object, signature)

    except (ValueError, IndexError, TypeError):
      raise securesystemslib.exceptions.CryptoError('The RSA signature could not'
        ' be verified.')

  else:
    raise securesystemslib.exceptions.UnknownMethodError(signature_method)

  return valid_signature





def create_rsa_encrypted_pem(private_key, passphrase):
  """
  <Purpose>
    Return a string in PEM format, where the private part of the RSA key is
    encrypted.  The private part of the RSA key is encrypted by the Triple
    Data Encryption Algorithm (3DES) and Cipher-block chaining (CBC) for the
    mode of operation.  Password-Based Key Derivation Function 1 (PBKF1) + MD5
    is used to strengthen 'passphrase'.

    https://en.wikipedia.org/wiki/Triple_DES
    https://en.wikipedia.org/wiki/PBKDF2

    >>> public, private = generate_rsa_public_and_private(2048)
    >>> passphrase = 'secret'
    >>> encrypted_pem = create_rsa_encrypted_pem(private, passphrase)
    >>> securesystemslib.formats.PEMRSA_SCHEMA.matches(encrypted_pem)
    True

  <Arguments>
    private_key:
      The private key string in PEM format.

    passphrase:
      The passphrase, or password, to encrypt the private part of the RSA
      key.  'passphrase' is not used directly as the encryption key, a stronger
      encryption key is derived from it.

  <Exceptions>
    securesystemslib.exceptions.FormatError, if the arguments are improperly formatted.

    securesystemslib.exceptions.CryptoError, if an RSA key in encrypted PEM format
    cannot be created.

    TypeError, 'private_key' is unset.

  <Side Effects>
    PyCrypto's Crypto.PublicKey.RSA.exportKey() called to perform the actual
    generation of the PEM-formatted output.

  <Returns>
    A string in PEM format, where the private RSA key is encrypted.
    Conforms to 'securesystemslib.formats.PEMRSA_SCHEMA'.
  """

  # Does 'private_key' have the correct format?
  # This check will ensure 'private_key' has the appropriate number
  # of objects and object types, and that all dict keys are properly named.
  # Raise 'securesystemslib.exceptions.FormatError' if the check fails.
  securesystemslib.formats.PEMRSA_SCHEMA.check_match(private_key)

  # Does 'passphrase' have the correct format?
  securesystemslib.formats.PASSWORD_SCHEMA.check_match(passphrase)

  # 'private_key' is in PEM format and unencrypted.  The extracted key will be
  # imported and converted to PyCrypto's RSA key object (i.e.,
  # Crypto.PublicKey.RSA).  Use PyCrypto's exportKey method, with a passphrase
  # specified, to create the string.  PyCrypto uses PBKDF1+MD5 to strengthen
  # 'passphrase', and 3DES with CBC mode for encryption.  'private_key' may
  # still be a NULL string after the 'securesystemslib.formats.PEMRSA_SCHEMA' (i.e.,
  # 'private_key' has variable size and can be an empty string.

  if len(private_key):
    try:
      rsa_key_object = Crypto.PublicKey.RSA.importKey(private_key)
      encrypted_pem = rsa_key_object.exportKey(format='PEM',
                                               passphrase=passphrase)

    except (ValueError, IndexError, TypeError) as e:
      raise securesystemslib.exceptions.CryptoError('An encrypted RSA key in PEM'
        ' format cannot be generated: ' + str(e))

  else:
    raise TypeError('The required private key is unset.')


  return encrypted_pem.decode()





def create_rsa_public_and_private_from_pem(pem, passphrase=None):
  """
  <Purpose>
    Generate public and private RSA keys from a PEM that can be encrypted.
    The public and private keys returned conform to
    'securesystemslib.formats.PEMRSA_SCHEMA' and have the form:

    '-----BEGIN RSA PUBLIC KEY----- ...'

    or

    '-----BEGIN RSA PRIVATE KEY----- ...'

    The public and private keys are returned as strings in PEM format.

    In case the private key part of 'pem' is encrypted PyCrypto's importKey
    method is passed a passphrase.  PyCrypto uses PBKDF1+MD5 to strengthen
    'passphrase', and 3DES with CBC mode for encryption/decryption.
    Alternatively, key data may be encrypted with AES-CTR-Mode and the
    passphrase strengthened with PBKDF2+SHA256.

    >>> public, private = generate_rsa_public_and_private(2048)
    >>> passphrase = 'secret'
    >>> encrypted_pem = create_rsa_encrypted_pem(private, passphrase)
    >>> returned_public, returned_private = \
    create_rsa_public_and_private_from_pem(encrypted_pem, passphrase)
    >>> securesystemslib.formats.PEMRSA_SCHEMA.matches(returned_public)
    True
    >>> securesystemslib.formats.PEMRSA_SCHEMA.matches(returned_private)
    True
    >>> public == returned_public
    True
    >>> private == returned_private
    True

  <Arguments>
    pem:
      A byte string in PEM format, where the private key can be encrypted.
      It has the form:

      '-----BEGIN RSA PRIVATE KEY-----\n
      Proc-Type: 4,ENCRYPTED\nDEK-Info: DES-EDE3-CBC ...'

    passphrase: (optional)
      The passphrase, or password, to decrypt the private part of the RSA
      key.  'passphrase' is not directly used as the encryption key, instead
      it is used to derive a stronger symmetric key.

  <Exceptions>
    securesystemslib.exceptions.FormatError, if the arguments are improperly
    formatted.

    securesystemslib.exceptions.CryptoError, if the public and private RSA keys
    cannot be generated from 'encrypted_pem', or exported in PEM format.

  <Side Effects>
    PyCrypto's 'Crypto.PublicKey.RSA.importKey()' called to perform the actual
    conversion from an (encrypted) RSA private key.

  <Returns>
    A (public, private) tuple containing the RSA keys in PEM format.
  """

  # Does 'pem' have the correct format?
  # This check will ensure 'pem' has the appropriate number
  # of objects and object types, and that all dict keys are properly named.
  # Raise 'securesystemslib.exceptions.FormatError' if the check fails.
  securesystemslib.formats.PEMRSA_SCHEMA.check_match(pem)

  # Does 'passphrase' have the correct format?
  securesystemslib.formats.PASSWORD_SCHEMA.check_match(passphrase)

  # Generate a PyCrypto key object from 'encrypted_pem'.  The generated
  # PyCrypto key contains the required export methods needed to generate the
  # PEM-formatted representations of the public and private RSA key.
  try:
    rsa_key_object = Crypto.PublicKey.RSA.importKey(pem, passphrase)

  # PyCrypto's expected exceptions:
  # "ValueError/IndexError/TypeError:  When the given key cannot be parsed
  # (possibly because the passphrase is wrong)."
  # If the passphrase is incorrect, PyCrypto returns: "RSA key format is not
  # supported".
  except (ValueError, IndexError, TypeError) as e:
    # Raise 'securesystemslib.exceptions.CryptoError' and PyCrypto's exception
    # message.  Avoid propogating PyCrypto's exception trace to avoid revealing
    # sensitive error.
    raise securesystemslib.exceptions.CryptoError('RSA (public, private) tuple'
      ' cannot be generated from the encrypted PEM string: ' + str(e))

  # Export the public and private halves of the PyCrypto RSA key object.  The
  # (public, private) tuple returned contains the public and private RSA keys
  # in PEM format, as strings.
  try:
    private = rsa_key_object.exportKey(format='PEM')
    rsa_pubkey = rsa_key_object.publickey()
    public = rsa_pubkey.exportKey(format='PEM')

  # PyCrypto raises 'ValueError' if the public or private keys cannot be
  # exported.  See 'Crypto.PublicKey.RSA'.  'ValueError' should not be raised
  # if the 'Crypto.PublicKey.RSA.importKey() call above passed.
  except (ValueError): #pragma: no cover
    raise securesystemslib.exceptions.CryptoError('The public and private keys'
      ' cannot be exported in PEM format.')

  return public.decode(), private.decode()





def encrypt_key(key_object, password):
  """
  <Purpose>
    Return a string containing 'key_object' in encrypted form. Encrypted strings
    may be safely saved to a file.  The corresponding decrypt_key() function can
    be applied to the encrypted string to restore the original key object.
    'key_object' is a TUF key (e.g., RSAKEY_SCHEMA, ED25519KEY_SCHEMA).  This
    function calls the PyCrypto library to perform the encryption and derive
    a suitable encryption key.

    Whereas an encrypted PEM file uses the Triple Data Encryption Algorithm
    (3DES), the Cipher-block chaining (CBC) mode of operation, and the Password
    Based Key Derivation Function 1 (PBKF1) + MD5 to strengthen 'password',
    encrypted TUF keys use AES-256-CTR-Mode and passwords strengthened with
    PBKDF2-HMAC-SHA256 (100K iterations by default, but may be overriden in
    'settings.PBKDF2_ITERATIONS' by the user).

    http://en.wikipedia.org/wiki/Advanced_Encryption_Standard
    http://en.wikipedia.org/wiki/CTR_mode#Counter_.28CTR.29
    https://en.wikipedia.org/wiki/PBKDF2

    >>> ed25519_key = {'keytype': 'ed25519', \
                       'keyid': \
          'd62247f817883f593cf6c66a5a55292488d457bcf638ae03207dbbba9dbe457d', \
                       'keyval': {'public': \
          '74addb5ad544a4306b34741bc1175a3613a8d7dc69ff64724243efdec0e301ad', \
                                  'private': \
          '1f26964cc8d4f7ee5f3c5da2fbb7ab35811169573ac367b860a537e47789f8c4'}}
    >>> passphrase = 'secret'
    >>> encrypted_key = encrypt_key(ed25519_key, passphrase)
    >>> securesystemslib.formats.ENCRYPTEDKEY_SCHEMA.matches(encrypted_key.encode('utf-8'))
    True

  <Arguments>
    key_object:
      The TUF key object that should contain the private portion of the ED25519
      key.

    password:
      The password, or passphrase, to encrypt the private part of the RSA
      key.  'password' is not used directly as the encryption key, a stronger
      encryption key is derived from it.

  <Exceptions>
    securesystemslib.exceptions.FormatError, if any of the arguments are improperly
    formatted or 'key_object' does not contain the private portion of the key.

    securesystemslib.exceptions.CryptoError, if an ED25519 key in encrypted TUF
    format cannot be created.

  <Side Effects>
    PyCrypto cryptographic operations called to perform the actual encryption of
    'key_object'.  'password' used to derive a suitable encryption key.

  <Returns>
    An encrypted string in 'securesystemslib.formats.ENCRYPTEDKEY_SCHEMA' format.
  """

  # Do the arguments have the correct format?
  # Ensure the arguments have the appropriate number of objects and object
  # types, and that all dict keys are properly named.
  # Raise 'securesystemslib.exceptions.FormatError' if the check fails.
  securesystemslib.formats.ANYKEY_SCHEMA.check_match(key_object)

  # Does 'password' have the correct format?
  securesystemslib.formats.PASSWORD_SCHEMA.check_match(password)

  # Ensure the private portion of the key is included in 'key_object'.
  if not key_object['keyval']['private']:
    raise securesystemslib.exceptions.FormatError('Key object does not contain'
      ' a private part.')

  # Derive a key (i.e., an appropriate encryption key and not the
  # user's password) from the given 'password'.  Strengthen 'password' with
  # PBKDF2-HMAC-SHA256 (100K iterations by default, but may be overriden in
  # 'settings.PBKDF2_ITERATIONS' by the user).
  salt, iterations, derived_key = _generate_derived_key(password)

  # Store the derived key info in a dictionary, the object expected
  # by the non-public _encrypt() routine.
  derived_key_information = {'salt': salt, 'iterations': iterations,
                             'derived_key': derived_key}

  # Convert the key object to json string format and encrypt it with the
  # derived key.
  encrypted_key = _encrypt(json.dumps(key_object), derived_key_information)

  return encrypted_key





def decrypt_key(encrypted_key, password):
  """
  <Purpose>
    Return a string containing 'encrypted_key' in non-encrypted form.
    The decrypt_key() function can be applied to the encrypted string to restore
    the original key object, a TUF key (e.g., RSAKEY_SCHEMA, ED25519KEY_SCHEMA).
    This function calls the appropriate cryptography module (e.g.,
    pycrypto_keys.py) to perform the decryption.

    Encrypted TUF keys use AES-256-CTR-Mode and passwords strengthened with
    PBKDF2-HMAC-SHA256 (100K iterations be default, but may be overriden in
    'settings.py' by the user).

    http://en.wikipedia.org/wiki/Advanced_Encryption_Standard
    http://en.wikipedia.org/wiki/CTR_mode#Counter_.28CTR.29
    https://en.wikipedia.org/wiki/PBKDF2

    >>> ed25519_key = {'keytype': 'ed25519', \
                       'keyid': \
          'd62247f817883f593cf6c66a5a55292488d457bcf638ae03207dbbba9dbe457d', \
                       'keyval': {'public': \
          '74addb5ad544a4306b34741bc1175a3613a8d7dc69ff64724243efdec0e301ad', \
                                  'private': \
          '1f26964cc8d4f7ee5f3c5da2fbb7ab35811169573ac367b860a537e47789f8c4'}}
    >>> passphrase = 'secret'
    >>> encrypted_key = encrypt_key(ed25519_key, passphrase)
    >>> decrypted_key = decrypt_key(encrypted_key.encode('utf-8'), passphrase)
    >>> securesystemslib.formats.ED25519KEY_SCHEMA.matches(decrypted_key)
    True
    >>> decrypted_key == ed25519_key
    True

  <Arguments>
    encrypted_key:
      An encrypted TUF key (additional data is also included, such as salt,
      number of password iterations used for the derived encryption key, etc)
      of the form 'securesystemslib.formats.ENCRYPTEDKEY_SCHEMA'.  'encrypted_key'
      should have been generated with encrypted_key().

    password:
      The password, or passphrase, to encrypt the private part of the RSA
      key.  'password' is not used directly as the encryption key, a stronger
      encryption key is derived from it.

  <Exceptions>
    securesystemslib.exceptions.FormatError, if the arguments are improperly
    formatted.

    securesystemslib.exceptions.CryptoError, if a TUF key cannot be decrypted from
    'encrypted_key'.

    securesystemslib.exceptions.Error, if a valid TUF key object is not found in
    'encrypted_key'.

  <Side Effects>
    The PyCrypto library called to perform the actual decryption of
    'encrypted_key'.  The key derivation data stored in 'encrypted_key' is used
    to re-derive the encryption/decryption key.

  <Returns>
    The decrypted key object in 'securesystemslib.formats.ANYKEY_SCHEMA' format.
  """

  # Do the arguments have the correct format?
  # Ensure the arguments have the appropriate number of objects and object
  # types, and that all dict keys are properly named.
  # Raise 'securesystemslib.exceptions.FormatError' if the check fails.
  securesystemslib.formats.ENCRYPTEDKEY_SCHEMA.check_match(encrypted_key)

  # Does 'password' have the correct format?
  securesystemslib.formats.PASSWORD_SCHEMA.check_match(password)

  # Decrypt 'encrypted_key', using 'password' (and additional key derivation
  # data like salts and password iterations) to re-derive the decryption key.
  json_data = _decrypt(encrypted_key, password)

  # Raise 'securesystemslib.exceptions.Error' if 'json_data' cannot be deserialized
  # to a valid 'securesystemslib.formats.ANYKEY_SCHEMA' key object.
  key_object = securesystemslib.util.load_json_string(json_data.decode())

  return key_object





def _generate_derived_key(password, salt=None, iterations=None):
  """
  Generate a derived key by feeding 'password' to the Password-Based Key
  Derivation Function (PBKDF2).  PyCrypto's PBKDF2 implementation is
  currently used.  'salt' may be specified so that a previous derived key
  may be regenerated.
  """

  if salt is None:
    salt = Crypto.Random.new().read(_SALT_SIZE)

  if iterations is None:
    iterations = _PBKDF2_ITERATIONS


  def pseudorandom_function(password, salt):
    """
    PyCrypto's PBKDF2() expects a callable function for its optional
    'prf' argument.  'prf' is set to HMAC-SHA1 (in PyCrypto's PBKDF2 function)
    by default.  'pseudorandom_function' instead sets 'prf' to HMAC-SHA256.
    """

    return Crypto.Hash.HMAC.new(password, salt, Crypto.Hash.SHA256).digest()


  # 'dkLen' is the desired key length.  'count' is the number of password
  # iterations performed by PBKDF2.  'prf' is a pseudorandom function, which
  # must be callable.
  derived_key = Crypto.Protocol.KDF.PBKDF2(password, salt,
                                           dkLen=_AES_KEY_SIZE,
                                           count=iterations,
                                           prf=pseudorandom_function)

  return salt, iterations, derived_key





def _encrypt(key_data, derived_key_information):
  """
  Encrypt 'key_data' using the Advanced Encryption Standard (AES-256) algorithm.
  'derived_key_information' should contain a key strengthened by PBKDF2.  The
  key size is 256 bits and AES's mode of operation is set to CTR (CounTeR Mode).
  The HMAC of the ciphertext is generated to ensure the ciphertext has not been
  modified.

  'key_data' is the JSON string representation of the key.  In the case
  of RSA keys, this format would be 'securesystemslib.formats.RSAKEY_SCHEMA':
  {'keytype': 'rsa',
   'keyval': {'public': '-----BEGIN RSA PUBLIC KEY----- ...',
              'private': '-----BEGIN RSA PRIVATE KEY----- ...'}}

  'derived_key_information' is a dictionary of the form:
    {'salt': '...',
     'derived_key': '...',
     'iterations': '...'}

  'securesystemslib.exceptions.CryptoError' raised if the encryption fails.
  """

  # Generate a random initialization vector (IV).  The 'iv' is treated as the
  # initial counter block to a stateful counter block function (i.e.,
  # PyCrypto's 'Crypto.Util.Counter').  The AES block cipher operates on 128-bit
  # blocks, so generate a random 16-byte initialization block.  PyCrypto expects
  # the initial value of the stateful counter to be an integer.
  # Follow the provably secure encrypt-then-MAC approach, which affords the
  # ability to verify ciphertext without needing to decrypt it and preventing
  # an attacker from feeding the block cipher malicious data.  Modes like GCM
  # provide both encryption and authentication, whereas CTR only provides
  # encryption.
  iv = Crypto.Random.new().read(16)
  stateful_counter_128bit_blocks = Crypto.Util.Counter.new(128,
                                      initial_value=int(binascii.hexlify(iv), 16))
  symmetric_key = derived_key_information['derived_key']
  aes_cipher = Crypto.Cipher.AES.new(symmetric_key,
                                     Crypto.Cipher.AES.MODE_CTR,
                                     counter=stateful_counter_128bit_blocks)

  # Use AES-256 to encrypt 'key_data'.  The key size determines how many cycle
  # repetitions are performed by AES, 14 cycles for 256-bit keys.
  try:
    ciphertext = aes_cipher.encrypt(key_data)

  # PyCrypto does not document the exceptions that may be raised or under what
  # circumstances.  PyCrypto example given is to call encrypt() without
  # checking for exceptions.  Avoid propogating the exception trace and only
  # raise 'securesystemslib.exceptions.CryptoError', along with the cause of
  # encryption failure.
  except (ValueError, IndexError, TypeError) as e:
    raise securesystemslib.exceptions.CryptoError('The key data cannot be'
      ' encrypted: ' + str(e))

  # Generate the hmac of the ciphertext to ensure it has not been modified.
  # The decryption routine may verify a ciphertext without having to perform
  # a decryption operation.
  salt = derived_key_information['salt']
  hmac_object = Crypto.Hash.HMAC.new(symmetric_key, ciphertext,
                                     Crypto.Hash.SHA256)
  hmac = hmac_object.hexdigest()

  # Store the number of PBKDF2 iterations used to derive the symmetric key so
  # that the decryption routine can regenerate the symmetric key successfully.
  # The pbkdf2 iterations are allowed to vary for the keys loaded and saved.
  iterations = derived_key_information['iterations']

  # Return the salt, iterations, hmac, initialization vector, and ciphertext
  # as a single string.  These five values are delimited by
  # '_ENCRYPTION_DELIMITER' to make extraction easier.  This delimiter is
  # arbitrarily chosen and should not occur in the hexadecimal representations
  # of the fields it is separating.
  return binascii.hexlify(salt).decode() + _ENCRYPTION_DELIMITER + \
         str(iterations) + _ENCRYPTION_DELIMITER + \
         hmac + _ENCRYPTION_DELIMITER + \
         binascii.hexlify(iv).decode() + _ENCRYPTION_DELIMITER + \
         binascii.hexlify(ciphertext).decode()





def _decrypt(file_contents, password):
  """
  The corresponding decryption routine for _encrypt().

  'securesystemslib.exceptions.CryptoError' raised if the decryption fails.
  """

  # Extract the salt, iterations, hmac, initialization vector, and ciphertext
  # from 'file_contents'.  These five values are delimited by
  # '_ENCRYPTION_DELIMITER'.  This delimiter is arbitrarily chosen and should
  # not occur in the hexadecimal representations of the fields it is
  # separating.  Raise 'securesystemslib.exceptions.CryptoError', if 'file_contents'
  # does not contains the expected data layout.
  try:
    salt, iterations, hmac, iv, ciphertext = \
      file_contents.split(_ENCRYPTION_DELIMITER)

  except ValueError:
    raise securesystemslib.exceptions.CryptoError('Invalid encrypted file.')

  # Ensure we have the expected raw data for the delimited cryptographic data.
  salt = binascii.unhexlify(salt.encode('utf-8'))
  iterations = int(iterations)
  iv = binascii.unhexlify(iv.encode('utf-8'))
  ciphertext = binascii.unhexlify(ciphertext.encode('utf-8'))

  # Generate derived key from 'password'.  The salt and iterations are
  # specified so that the expected derived key is regenerated correctly.
  # Discard the old "salt" and "iterations" values, as we only need the old
  # derived key.
  junk_old_salt, junk_old_iterations, derived_key = \
    _generate_derived_key(password, salt, iterations)

  # Verify the hmac to ensure the ciphertext is valid and has not been altered.
  # See the encryption routine for why we use the encrypt-then-MAC approach.
  generated_hmac_object = Crypto.Hash.HMAC.new(derived_key, ciphertext,
                                               Crypto.Hash.SHA256)
  generated_hmac = generated_hmac_object.hexdigest()

  if not securesystemslib.util.digests_are_equal(generated_hmac, hmac):
    raise securesystemslib.exceptions.CryptoError('Decryption failed.')

  # The following decryption routine assumes 'ciphertext' was encrypted with
  # AES-256.
  stateful_counter_128bit_blocks = Crypto.Util.Counter.new(128,
                                      initial_value=int(binascii.hexlify(iv), 16))
  aes_cipher = Crypto.Cipher.AES.new(derived_key,
                                     Crypto.Cipher.AES.MODE_CTR,
                                     counter=stateful_counter_128bit_blocks)
  try:
    key_plaintext = aes_cipher.decrypt(ciphertext)

  # PyCrypto does not document the exceptions that may be raised or under what
  # circumstances.  PyCrypto example given is to call decrypt() without
  # checking for exceptions.  Avoid propogating the exception trace and only
  # raise 'securesystemslib.exceptions.CryptoError', along with the cause of
  # decryption failure.  Note: decryption failure, due to malicious ciphertext,
  # should not occur here if the hmac check above passed.
  except (ValueError, IndexError, TypeError) as e: # pragma: no cover
    raise securesystemslib.exceptions.CryptoError('Decryption failed: ' + str(e))

  return key_plaintext



if __name__ == '__main__':
  # The interactive sessions of the documentation strings can
  # be tested by running 'pycrypto_keys.py' as a standalone module:
  # $ python pycrypto_keys.py
  import doctest
  doctest.testmod()



#!/usr/bin/env python

"""
<Program Name>
  interface.py

<Author>
  Vladimir Diaz <vladimir.v.diaz@gmail.com>

<Started>
  January 5, 2017.

<Copyright>
  See LICENSE for licensing information.

<Purpose>
  Provide an interface to the cryptography funtions available in
  securesystemslib.  The interface can be used with the Python interpreter in
  interactive mode, or imported directly into a Python module.  See
  'securesystemslib/README' for the complete guide to using 'interface.py'.
"""

# Help with Python 3 compatibility, where the print statement is a function, an
# implicit relative import is invalid, and the '/' operator performs true
# division.  Example:  print 'hello world' raises a 'SyntaxError' exception.
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

import os
import errno
import sys
import time
import datetime
import getpass
import logging
import tempfile
import shutil
import json
import gzip
import random

import securesystemslib.formats
import securesystemslib.settings
import securesystemslib.util
import securesystemslib.keys

import six

# See 'log.py' to learn how logging is handled in securesystemslib.
logger = logging.getLogger('securesystemslib_interface')


# Recommended RSA key sizes:
# http://www.emc.com/emc-plus/rsa-labs/historical/twirl-and-rsa-key-size.htm#table1
# According to the document above, revised May 6, 2003, RSA keys of
# size 3072 provide security through 2031 and beyond. 2048-bit keys
# are the recommended minimum and are good from the present through 2030.
DEFAULT_RSA_KEY_BITS = 3072

# Supported key types.
SUPPORTED_KEY_TYPES = ['rsa', 'ed25519']



def _prompt(message, result_type=str):
  """
    Non-public function that prompts the user for input by logging 'message',
    converting the input to 'result_type', and returning the value to the
    caller.
  """

  return result_type(six.moves.input(message))





def _get_password(prompt='Password: ', confirm=False):
  """
    Non-public function that returns the password entered by the user.  If
    'confirm' is True, the user is asked to enter the previously entered
    password once again.  If they match, the password is returned to the caller.
  """

  while True:
    # getpass() prompts the user for a password without echoing
    # the user input.
    password = getpass.getpass(prompt, sys.stderr)

    if not confirm:
      return password
    password2 = getpass.getpass('Confirm: ', sys.stderr)

    if password == password2:
      return password

    else:
      print('Mismatch; try again.')





def generate_and_write_rsa_keypair(filepath, bits=DEFAULT_RSA_KEY_BITS,
                                   password=None):
  """
  <Purpose>
    Generate an RSA key file, create an encrypted PEM string (using 'password'
    as the pass phrase), and store it in 'filepath'.  The public key portion of
    the generated RSA key is stored in <'filepath'>.pub.  Which cryptography
    library performs the cryptographic decryption is determined by the string
    set in 'settings.RSA_CRYPTO_LIBRARY'.  PyCrypto currently supported.  The
    PEM private key is encrypted with 3DES and CBC the mode of operation.  The
    password is strengthened with PBKDF1-MD5.

  <Arguments>
    filepath:
      The public and private key files are saved to <filepath>.pub, <filepath>,
      respectively.

    bits:
      The number of bits of the generated RSA key.

    password:
      The password used to encrypt 'filepath'.

  <Exceptions>
    securesystemslib.exceptions.FormatError, if the arguments are improperly
    formatted.

  <Side Effects>
    Writes key files to '<filepath>' and '<filepath>.pub'.

  <Returns>
    None.
  """

  # Do the arguments have the correct format?
  # This check ensures arguments have the appropriate number of
  # objects and object types, and that all dict keys are properly named.
  # Raise 'securesystemslib.exceptions.FormatError' if there is a mismatch.
  securesystemslib.formats.PATH_SCHEMA.check_match(filepath)

  # Does 'bits' have the correct format?
  securesystemslib.formats.RSAKEYBITS_SCHEMA.check_match(bits)

  # If the caller does not provide a password argument, prompt for one.
  if password is None: # pragma: no cover
    message = 'Enter a password for the RSA key file: '
    password = _get_password(message, confirm=True)

  # Does 'password' have the correct format?
  securesystemslib.formats.PASSWORD_SCHEMA.check_match(password)

  #  Generate public and private RSA keys, encrypted the private portion
  # and store them in PEM format.
  rsa_key = securesystemslib.keys.generate_rsa_key(bits)
  public = rsa_key['keyval']['public']
  private = rsa_key['keyval']['private']
  encrypted_pem = securesystemslib.keys.create_rsa_encrypted_pem(private, password)

  # Write public key (i.e., 'public', which is in PEM format) to
  # '<filepath>.pub'.  If the parent directory of filepath does not exist,
  # create it (and all its parent directories, if necessary).
  securesystemslib.util.ensure_parent_dir(filepath)

  # Create a tempororary file, write the contents of the public key, and move
  # to final destination.
  file_object = securesystemslib.util.TempFile()
  file_object.write(public.encode('utf-8'))

  # The temporary file is closed after the final move.
  file_object.move(filepath + '.pub')

  # Write the private key in encrypted PEM format to '<filepath>'.
  # Unlike the public key file, the private key does not have a file
  # extension.
  file_object = securesystemslib.util.TempFile()
  file_object.write(encrypted_pem.encode('utf-8'))
  file_object.move(filepath)





def import_rsa_privatekey_from_file(filepath, password=None):
  """
  <Purpose>
    Import the encrypted PEM file in 'filepath', decrypt it, and return the key
    object in 'securesystemslib.formats.RSAKEY_SCHEMA' format.

    Which cryptography library performs the cryptographic decryption is
    determined by the string set in 'settings.RSA_CRYPTO_LIBRARY'.  PyCrypto
    currently supported.

    The PEM private key is encrypted with 3DES and CBC the mode of operation.
    The password is strengthened with PBKDF1-MD5.

  <Arguments>
    filepath:
      <filepath> file, an RSA encrypted PEM file.  Unlike the public RSA PEM
      key file, 'filepath' does not have an extension.

    password:
      The passphrase to decrypt 'filepath'.

  <Exceptions>
    securesystemslib.exceptions.FormatError, if the arguments are improperly
    formatted.

    securesystemslib.exceptions.CryptoError, if 'filepath' is not a valid
    encrypted key file.

  <Side Effects>
    The contents of 'filepath' is read, decrypted, and the key stored.

  <Returns>
    An RSA key object, conformant to 'securesystemslib.formats.RSAKEY_SCHEMA'.
  """

  # Does 'filepath' have the correct format?
  # Ensure the arguments have the appropriate number of objects and object
  # types, and that all dict keys are properly named.
  # Raise 'securesystemslib.exceptions.FormatError' if there is a mismatch.
  securesystemslib.formats.PATH_SCHEMA.check_match(filepath)

  # If the caller does not provide a password argument, prompt for one.
  # Password confirmation disabled here, which should ideally happen only
  # when creating encrypted key files (i.e., improve usability).
  if password is None: # pragma: no cover
    message = 'Enter a password for the encrypted RSA file: '
    password = _get_password(message, confirm=False)

  # Does 'password' have the correct format?
  securesystemslib.formats.PASSWORD_SCHEMA.check_match(password)

  encrypted_pem = None

  # Read the contents of 'filepath' that should be an encrypted PEM.
  with open(filepath, 'rb') as file_object:
    encrypted_pem = file_object.read().decode('utf-8')

  # Convert 'encrypted_pem' to 'securesystemslib.formats.RSAKEY_SCHEMA' format.
  # Raise 'securesystemslib.exceptions.CryptoError' if 'encrypted_pem' is
  # invalid.
  rsa_key = securesystemslib.keys.import_rsakey_from_private_pem(encrypted_pem, password)

  return rsa_key





def import_rsa_publickey_from_file(filepath):
  """
  <Purpose>
    Import the RSA key stored in 'filepath'.  The key object returned is in the
    format 'securesystemslib.formats.RSAKEY_SCHEMA'.  If the RSA PEM in
    'filepath' contains a private key, it is discarded.

    Which cryptography library performs the cryptographic decryption is
    determined by the string set in 'settings.RSA_CRYPTO_LIBRARY'.  PyCrypto
    currently supported.  If the RSA PEM in 'filepath' contains a private key,
    it is discarded.

  <Arguments>
    filepath:
      <filepath>.pub file, an RSA PEM file.

  <Exceptions>
    securesystemslib.exceptions.FormatError, if 'filepath' is improperly formatted.

    securesystemslib.exceptions.Error, if a valid RSA key object cannot be
    generated.  This may be caused by an improperly formatted PEM file.

  <Side Effects>
    'filepath' is read and its contents extracted.

  <Returns>
    An RSA key object conformant to 'securesystemslib.formats.RSAKEY_SCHEMA'.
  """

  # Does 'filepath' have the correct format?
  # Ensure the arguments have the appropriate number of objects and object
  # types, and that all dict keys are properly named.
  # Raise 'securesystemslib.exceptions.FormatError' if there is a mismatch.
  securesystemslib.formats.PATH_SCHEMA.check_match(filepath)

  # Read the contents of the key file that should be in PEM format and contains
  # the public portion of the RSA key.
  with open(filepath, 'rb') as file_object:
    rsa_pubkey_pem = file_object.read().decode('utf-8')

  # Convert 'rsa_pubkey_pem' to 'securesystemslib.formats.RSAKEY_SCHEMA' format.
  try:
    rsakey_dict = securesystemslib.keys.import_rsakey_from_public_pem(rsa_pubkey_pem)

  except securesystemslib.exceptions.FormatError as e:
    raise securesystemslib.exceptions.Error('Cannot import improperly formatted'
      ' PEM file.' + repr(str(e)))

  return rsakey_dict





def generate_and_write_ed25519_keypair(filepath, password=None):
  """
  <Purpose>
    Generate an Ed25519 key file, create an encrypted key (using 'password'
    as the pass phrase), and store it in 'filepath'.  The public key portion of
    the generated ED25519 key is stored in <'filepath'>.pub.  Which cryptography
    library performs the cryptographic decryption is determined by the string
    set in 'settings.ED25519_CRYPTO_LIBRARY'.

    PyCrypto currently supported.  The Ed25519 private key is encrypted with
    AES-256 and CTR the mode of operation.  The password is strengthened with
    PBKDF2-HMAC-SHA256.

  <Arguments>
    filepath:
      The public and private key files are saved to <filepath>.pub and
      <filepath>, respectively.

    password:
      The password, or passphrase, to encrypt the private portion of the
      generated ed25519 key.  A symmetric encryption key is derived from
      'password', so it is not directly used.

  <Exceptions>
    securesystemslib.exceptions.FormatError, if the arguments are improperly
    formatted.

    securesystemslib.exceptions.CryptoError, if 'filepath' cannot be encrypted.

    securesystemslib.exceptions.UnsupportedLibraryError, if 'filepath' cannot be
    encrypted due to an invalid configuration setting (i.e., invalid
    'settings.py' setting).

  <Side Effects>
    Writes key files to '<filepath>' and '<filepath>.pub'.

  <Returns>
    None.
  """

  # Does 'filepath' have the correct format?
  # Ensure the arguments have the appropriate number of objects and object
  # types, and that all dict keys are properly named.
  # Raise 'securesystemslib.exceptions.FormatError' if there is a mismatch.
  securesystemslib.formats.PATH_SCHEMA.check_match(filepath)

  # If the caller does not provide a password argument, prompt for one.
  if password is None: # pragma: no cover
    message = 'Enter a password for the Ed25519 key: '
    password = _get_password(message, confirm=True)

  # Does 'password' have the correct format?
  securesystemslib.formats.PASSWORD_SCHEMA.check_match(password)

  # Generate a new ED25519 key object and encrypt it.  The cryptography library
  # used is determined by the user, or by default (set in
  # 'settings.ED25519_CRYPTO_LIBRARY').  Raise
  # 'securesystemslib.exceptions.CryptoError' or
  # 'securesystemslib.exceptions.UnsupportedLibraryError', if 'ed25519_key'
  # cannot be encrypted.
  ed25519_key = securesystemslib.keys.generate_ed25519_key()
  encrypted_key = securesystemslib.keys.encrypt_key(ed25519_key, password)

  # ed25519 public key file contents in metadata format (i.e., does not include
  # the keyid portion).
  keytype = ed25519_key['keytype']
  keyval = ed25519_key['keyval']
  ed25519key_metadata_format = \
    securesystemslib.keys.format_keyval_to_metadata(keytype, keyval, private=False)

  # Write the public key, conformant to 'securesystemslib.formats.KEY_SCHEMA', to
  # '<filepath>.pub'.
  securesystemslib.util.ensure_parent_dir(filepath)

  # Create a tempororary file, write the contents of the public key, and move
  # to final destination.
  file_object = securesystemslib.util.TempFile()
  file_object.write(json.dumps(ed25519key_metadata_format).encode('utf-8'))

  # The temporary file is closed after the final move.
  file_object.move(filepath + '.pub')

  # Write the encrypted key string, conformant to
  # 'securesystemslib.formats.ENCRYPTEDKEY_SCHEMA', to '<filepath>'.
  file_object = securesystemslib.util.TempFile()
  file_object.write(encrypted_key.encode('utf-8'))
  file_object.move(filepath)





def import_ed25519_publickey_from_file(filepath):
  """
  <Purpose>
    Load the ED25519 public key object (conformant to
    'securesystemslib.formats.KEY_SCHEMA') stored in 'filepath'.  Return
    'filepath' in securesystemslib.formats.ED25519KEY_SCHEMA format.

    If the key object in 'filepath' contains a private key, it is discarded.

  <Arguments>
    filepath:
      <filepath>.pub file, a public key file.

  <Exceptions>
    securesystemslib.exceptions.FormatError, if 'filepath' is improperly
    formatted or is an unexpected key type.

  <Side Effects>
    The contents of 'filepath' is read and saved.

  <Returns>
    An ED25519 key object conformant to
    'securesystemslib.formats.ED25519KEY_SCHEMA'.
  """

  # Does 'filepath' have the correct format?
  # Ensure the arguments have the appropriate number of objects and object
  # types, and that all dict keys are properly named.
  # Raise 'securesystemslib.exceptions.FormatError' if there is a mismatch.
  securesystemslib.formats.PATH_SCHEMA.check_match(filepath)

  # ED25519 key objects are saved in json and metadata format.  Return the
  # loaded key object in securesystemslib.formats.ED25519KEY_SCHEMA' format that
  # also includes the keyid.
  ed25519_key_metadata = securesystemslib.util.load_json_file(filepath)
  ed25519_key, junk = securesystemslib.keys.format_metadata_to_key(ed25519_key_metadata)

  # Raise an exception if an unexpected key type is imported.  Redundant
  # validation of 'keytype'.  'securesystemslib.keys.format_metadata_to_key()'
  # should have fully validated 'ed25519_key_metadata'.
  if ed25519_key['keytype'] != 'ed25519': # pragma: no cover
    message = 'Invalid key type loaded: ' + repr(ed25519_key['keytype'])
    raise securesystemslib.exceptions.FormatError(message)

  return ed25519_key





def import_ed25519_privatekey_from_file(filepath, password=None):
  """
  <Purpose>
    Import the encrypted ed25519 key file in 'filepath', decrypt it, and
    return the key object in 'securesystemslib.formats.ED25519KEY_SCHEMA' format.

    Which cryptography library performs the cryptographic decryption is
    determined by the string set in 'settings.ED25519_CRYPTO_LIBRARY'.  PyCrypto
    currently supported.

    The private key (may also contain the public part) is encrypted with AES
    256 and CTR the mode of operation.  The password is strengthened with
    PBKDF2-HMAC-SHA256.

  <Arguments>
    filepath:
      <filepath> file, an RSA encrypted key file.

    password:
      The password, or passphrase, to import the private key (i.e., the
      encrypted key file 'filepath' must be decrypted before the ed25519 key
      object can be returned.

  <Exceptions>
    securesystemslib.exceptions.FormatError, if the arguments are improperly
    formatted or the imported key object contains an invalid key type (i.e.,
    not 'ed25519').

    securesystemslib.exceptions.CryptoError, if 'filepath' cannot be decrypted.

    securesystemslib.exceptions.UnsupportedLibraryError, if 'filepath' cannot be
    decrypted due to an invalid configuration setting (i.e., invalid
    'settings.py' setting).

  <Side Effects>
    'password' is used to decrypt the 'filepath' key file.

  <Returns>
    An ed25519 key object of the form: 'securesystemslib.formats.ED25519KEY_SCHEMA'.
  """

  # Does 'filepath' have the correct format?
  # Ensure the arguments have the appropriate number of objects and object
  # types, and that all dict keys are properly named.
  # Raise 'securesystemslib.exceptions.FormatError' if there is a mismatch.
  securesystemslib.formats.PATH_SCHEMA.check_match(filepath)

  # If the caller does not provide a password argument, prompt for one.
  # Password confirmation disabled here, which should ideally happen only
  # when creating encrypted key files (i.e., improve usability).
  if password is None: # pragma: no cover
    message = 'Enter a password for the encrypted Ed25519 key: '
    password = _get_password(message, confirm=False)

  # Does 'password' have the correct format?
  securesystemslib.formats.PASSWORD_SCHEMA.check_match(password)

  # Store the encrypted contents of 'filepath' prior to calling the decryption
  # routine.
  encrypted_key = None

  with open(filepath, 'rb') as file_object:
    encrypted_key = file_object.read()

  # Decrypt the loaded key file, calling the appropriate cryptography library
  # (i.e., set by the user) and generating the derived encryption key from
  # 'password'.  Raise 'securesystemslib.exceptions.CryptoError' or
  # 'securesystemslib.exceptions.UnsupportedLibraryError' if the decryption
  # fails.
  key_object = securesystemslib.keys.decrypt_key(encrypted_key.decode('utf-8'), password)

  # Raise an exception if an unexpected key type is imported.
  if key_object['keytype'] != 'ed25519':
    message = 'Invalid key type loaded: ' + repr(key_object['keytype'])
    raise securesystemslib.exceptions.FormatError(message)

  # Add "keyid_hash_algorithms" so that equal ed25519 keys with
  # different keyids can be associated using supported keyid_hash_algorithms.
  key_object['keyid_hash_algorithms'] = \
      securesystemslib.settings.HASH_ALGORITHMS

  return key_object





def generate_and_write_ecdsa_keypair(filepath, password=None):
  """
  <Purpose>
    Generate an ECDSA key file, create an encrypted key (using 'password' as
    the pass phrase), and store it in 'filepath'.  The public key portion of
    the generated ECDSA key is stored in <'filepath'>.pub.  Which cryptography
    library performs the cryptographic decryption is determined by the string
    set in 'settings.ECDSA_CRYPTO_LIBRARY'.  'cryptography' library currently
    supported.  The private key is encrypted according to 'cryptography's
    approach: "Encrypt using the best available encryption for a given key's
    backend. This is a curated encryption choice and the algorithm may
    change over time."

  <Arguments>
    filepath:
      The public and private key files are saved to <filepath>.pub and
      <filepath>, respectively.

    password:
      The password, or passphrase, to encrypt the private portion of the
      generated ECDSA key.  A symmetric encryption key is derived from
      'password', so it is not directly used.

  <Exceptions>
    securesystemslib.exceptions.FormatError, if the arguments are improperly
    formatted.

    securesystemslib.exceptions.CryptoError, if 'filepath' cannot be encrypted.

    securesystemslib.exceptions.UnsupportedLibraryError, if 'filepath' cannot be
    encrypted due to an invalid configuration setting (i.e., invalid
    'settings.py' setting).

  <Side Effects>
    Writes key files to '<filepath>' and '<filepath>.pub'.

  <Returns>
    None.
  """

  # Does 'filepath' have the correct format?
  # Ensure the arguments have the appropriate number of objects and object
  # types, and that all dict keys are properly named.
  # Raise 'securesystemslib.exceptions.FormatError' if there is a mismatch.
  securesystemslib.formats.PATH_SCHEMA.check_match(filepath)

  # If the caller does not provide a password argument, prompt for one.
  if password is None: # pragma: no cover
    message = 'Enter a password for the ECDSA key: '
    password = _get_password(message, confirm=True)

  # Does 'password' have the correct format?
  securesystemslib.formats.PASSWORD_SCHEMA.check_match(password)

  # Generate a new ECDSA key object and encrypt it.  The cryptography library
  # used is determined by the user, or by default (set in
  # 'settings.ECDSA_CRYPTO_LIBRARY').  Raise
  # 'securesystemslib.exceptions.CryptoError' or
  # 'securesystemslib.exceptions.UnsupportedLibraryError', if 'ed25519_key'
  # cannot be encrypted.
  ecdsa_key = securesystemslib.keys.generate_ecdsa_key()
  encrypted_key = securesystemslib.keys.encrypt_key(ecdsa_key, password)

  # ECDSA public key file contents in metadata format (i.e., does not include
  # the keyid portion).
  keytype = ecdsa_key['keytype']
  keyval = ecdsa_key['keyval']
  ecdsakey_metadata_format = \
    securesystemslib.keys.format_keyval_to_metadata(keytype, keyval, private=False)

  # Write the public key, conformant to 'securesystemslib.formats.KEY_SCHEMA', to
  # '<filepath>.pub'.
  securesystemslib.util.ensure_parent_dir(filepath)

  # Create a tempororary file, write the contents of the public key, and move
  # to final destination.
  file_object = securesystemslib.util.TempFile()
  file_object.write(json.dumps(ecdsakey_metadata_format).encode('utf-8'))

  # The temporary file is closed after the final move.
  file_object.move(filepath + '.pub')

  # Write the encrypted key string, conformant to
  # 'securesystemslib.formats.ENCRYPTEDKEY_SCHEMA', to '<filepath>'.
  file_object = securesystemslib.util.TempFile()
  file_object.write(encrypted_key.encode('utf-8'))
  file_object.move(filepath)





def import_ecdsa_publickey_from_file(filepath):
  """
  <Purpose>
    Load the ECDSA public key object (conformant to
    'securesystemslib.formats.KEY_SCHEMA') stored in 'filepath'.  Return
    'filepath' in securesystemslib.formats.ECDSAKEY_SCHEMA format.

    If the key object in 'filepath' contains a private key, it is discarded.

  <Arguments>
    filepath:
      <filepath>.pub file, a public key file.

  <Exceptions>
    securesystemslib.exceptions.FormatError, if 'filepath' is improperly
    formatted or is an unexpected key type.

  <Side Effects>
    The contents of 'filepath' is read and saved.

  <Returns>
    An ECDSA key object conformant to
    'securesystemslib.formats.ECDSAKEY_SCHEMA'.
  """

  # Does 'filepath' have the correct format?
  # Ensure the arguments have the appropriate number of objects and object
  # types, and that all dict keys are properly named.
  # Raise 'securesystemslib.exceptions.FormatError' if there is a mismatch.
  securesystemslib.formats.PATH_SCHEMA.check_match(filepath)

  # ECDSA key objects are saved in json and metadata format.  Return the
  # loaded key object in securesystemslib.formats.ECDSAKEY_SCHEMA' format that
  # also includes the keyid.
  ecdsa_key_metadata = securesystemslib.util.load_json_file(filepath)
  ecdsa_key, junk = securesystemslib.keys.format_metadata_to_key(ecdsa_key_metadata)

  # Raise an exception if an unexpected key type is imported.  Redundant
  # validation of 'keytype'.  'securesystemslib.keys.format_metadata_to_key()'
  # should have fully validated 'ecdsa_key_metadata'.
  if ecdsa_key['keytype'] != 'ecdsa-sha2-nistp256': # pragma: no cover
    message = 'Invalid key type loaded: ' + repr(ecdsa_key['keytype'])
    raise securesystemslib.exceptions.FormatError(message)

  return ecdsa_key





def import_ecdsa_privatekey_from_file(filepath, password=None):
  """
  <Purpose>
    Import the encrypted ECDSA key file in 'filepath', decrypt it, and return
    the key object in 'securesystemslib.formats.ECDSAKEY_SCHEMA' format.

    Which cryptography library performs the cryptographic decryption is
    determined by the string set in 'settings.ECDSA_CRYPTO_LIBRARY'.
    currently supported.  The 'cryptography' library currently supported.

  <Arguments>
    filepath:
      <filepath> file, an ECDSA encrypted key file.

    password:
      The password, or passphrase, to import the private key (i.e., the
      encrypted key file 'filepath' must be decrypted before the ECDSA key
      object can be returned.

  <Exceptions>
    securesystemslib.exceptions.FormatError, if the arguments are improperly
    formatted or the imported key object contains an invalid key type (i.e.,
    not 'ecdsa-sha2-nistp256').

    securesystemslib.exceptions.CryptoError, if 'filepath' cannot be decrypted.

    securesystemslib.exceptions.UnsupportedLibraryError, if 'filepath' cannot be
    decrypted due to an invalid configuration setting (i.e., invalid
    'settings.py' setting).

  <Side Effects>
    'password' is used to decrypt the 'filepath' key file.

  <Returns>
    An ECDSA key object of the form: 'securesystemslib.formats.ECDSAKEY_SCHEMA'.
  """

  # Does 'filepath' have the correct format?
  # Ensure the arguments have the appropriate number of objects and object
  # types, and that all dict keys are properly named.
  # Raise 'securesystemslib.exceptions.FormatError' if there is a mismatch.
  securesystemslib.formats.PATH_SCHEMA.check_match(filepath)

  # If the caller does not provide a password argument, prompt for one.
  # Password confirmation disabled here, which should ideally happen only
  # when creating encrypted key files (i.e., improve usability).
  if password is None: # pragma: no cover
    message = 'Enter a password for the encrypted ECDSA key: '
    password = _get_password(message, confirm=False)

  # Does 'password' have the correct format?
  securesystemslib.formats.PASSWORD_SCHEMA.check_match(password)

  # Store the encrypted contents of 'filepath' prior to calling the decryption
  # routine.
  encrypted_key = None

  with open(filepath, 'rb') as file_object:
    encrypted_key = file_object.read()

  # Decrypt the loaded key file, calling the appropriate cryptography library
  # (i.e., set by the user) and generating the derived encryption key from
  # 'password'.  Raise 'securesystemslib.exceptions.CryptoError' or
  # 'securesystemslib.exceptions.UnsupportedLibraryError' if the decryption
  # fails.
  key_object = securesystemslib.keys.decrypt_key(encrypted_key.decode('utf-8'), password)

  # Raise an exception if an unexpected key type is imported.
  if key_object['keytype'] != 'ecdsa-sha2-nistp256':
    message = 'Invalid key type loaded: ' + repr(key_object['keytype'])
    raise securesystemslib.exceptions.FormatError(message)

  # Add "keyid_hash_algorithms" so that equal ecdsa keys with different keyids
  # can be associated using supported keyid_hash_algorithms.
  key_object['keyid_hash_algorithms'] = \
      securesystemslib.settings.HASH_ALGORITHMS

  return key_object



if __name__ == '__main__':
  # The interactive sessions of the documentation strings can
  # be tested by running interface.py as a standalone module:
  # $ python interface.py.
  import doctest
  doctest.testmod()



#!/usr/bin/env python2
"""
<Program Name>
  hash.py

<Author>
  Vladimir Diaz <vladimir.v.diaz@gmail.com>

<Started>
  February 28, 2012.  Based on a previous version of this module.

<Copyright>
  See LICENSE for licensing information.

<Purpose>
  Support multiple implementations of secure hash and message digest
  algorithms. Any hash-related routines that TUF requires should be
  located in this module.  Ensuring that a secure hash algorithm is
  available to TUF, simplifying the creation of digest objects, and
  providing a central location for hash routines are the main goals
  of this module.  Support routines implemented include functions to
  create digest objects given a filename or file object.  Hashlib and PyCrypto
  hash algorithms currently supported.
"""

# Help with Python 3 compatibility, where the print statement is a function, an
# implicit relative import is invalid, and the '/' operator performs true
# division.  Example:  print 'hello world' raises a 'SyntaxError' exception.
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

import logging

import six

import securesystemslib.exceptions

# Import securesystemslib logger to log warning messages.
logger = logging.getLogger('securesystemslib.hash')

# The list of hash libraries imported successfully.
_supported_libraries = []

# Hash libraries currently supported by securesystemslib.hash.
_SUPPORTED_LIB_LIST = ['hashlib', 'pycrypto']

# Let's try importing the pycrypto hash algorithms.  Pycrypto will
# not be added to the supported list of libraries if the specified
# hash algorithms below cannot all be imported.
try:
  from Crypto.Hash import MD5
  from Crypto.Hash import SHA
  from Crypto.Hash import SHA224
  from Crypto.Hash import SHA256
  from Crypto.Hash import SHA384
  from Crypto.Hash import SHA512
  _supported_libraries.append('pycrypto')

except ImportError: # pragma: no cover
  logger.debug('Pycrypto hash algorithms could not be imported.  '
              'Supported libraries: '+str(_SUPPORTED_LIB_LIST))

  pass

# Python <=2.4 does not have the hashlib module by default.
# Let's try importing hashlib and adding it to our supported list.
try:
  import hashlib
  _supported_libraries.append('hashlib')

except ImportError: # pragma: no cover
  logger.debug('Hashlib could not be imported.  '
              'Supported libraries: '+str(_SUPPORTED_LIB_LIST))
  pass

# Were we able to import any hash libraries?
if not _supported_libraries: # pragma: no cover
  # This is fatal, we'll have no way of generating hashes.
  raise securesystemslib.exceptions.Error('Unable to import a hash library from the '
                  'following supported list: '+str(_SUPPORTED_LIB_LIST))


_DEFAULT_HASH_ALGORITHM = 'sha256'
_DEFAULT_HASH_LIBRARY = 'hashlib'





def digest(algorithm=_DEFAULT_HASH_ALGORITHM,
           hash_library=_DEFAULT_HASH_LIBRARY):
  """
  <Purpose>
    Provide the caller with the ability to create
    digest objects without having to worry about hash
    library availability or which library to use.
    The caller also has the option of specifying which
    hash algorithm and/or library to use.

    # Creation of a digest object using defaults
    # or by specifying hash algorithm and library.
    digest_object = securesystemslib.hash.digest()
    digest_object = securesystemslib.hash.digest('sha384')
    digest_object = securesystemslib.hash.digest('pycrypto')

    # The expected interface for digest objects.
    digest_object.digest_size
    digest_object.hexdigest()
    digest_object.update('data')
    digest_object.digest()

    # Added hash routines by this module.
    digest_object = securesystemslib.hash.digest_fileobject(file_object)
    digest_object = securesystemslib.hash.digest_filename(filename)

  <Arguments>
    algorithm:
      The hash algorithm (e.g., md5, sha1, sha256).

    hash_library:
      The library providing the hash algorithms
      (e.g., pycrypto, hashlib).

  <Exceptions>
    securesystemslib.exceptions.UnsupportedAlgorithmError
    securesystemslib.exceptions.UnsupportedLibraryError

  <Side Effects>
    None.

  <Returns>
    Digest object (e.g., hashlib.new(algorithm) or
    algorithm.new() # pycrypto).
  """

  # Was a hashlib digest object requested and is it supported?
  # If so, return the digest object.
  if hash_library == 'hashlib' and hash_library in _supported_libraries:
    try:
      return hashlib.new(algorithm)

    except ValueError:
      raise securesystemslib.exceptions.UnsupportedAlgorithmError(algorithm)

  # Was a pycrypto digest object requested and is it supported?
  elif hash_library == 'pycrypto' and hash_library in _supported_libraries:
    # Pycrypto does not offer a comparable hashlib.new(hashname).
    # Let's first check the 'algorithm' argument before returning
    # the correct pycrypto digest object using pycrypto's object construction.
    if algorithm == 'md5':
      return MD5.new()
    elif algorithm == 'sha1':
      return SHA.new()
    elif algorithm == 'sha224':
      return SHA224.new()
    elif algorithm == 'sha256':
      return SHA256.new()
    elif algorithm == 'sha384':
      return SHA384.new()
    elif algorithm == 'sha512':
      return SHA512.new()
    else:
      raise securesystemslib.exceptions.UnsupportedAlgorithmError(algorithm)

  # The requested hash library is not supported.
  else:
    raise securesystemslib.exceptions.UnsupportedLibraryError('Unsupported library requested.  '
                    'Supported hash libraries: '+str(_SUPPORTED_LIB_LIST))





def digest_fileobject(file_object, algorithm=_DEFAULT_HASH_ALGORITHM,
                      hash_library=_DEFAULT_HASH_LIBRARY):
  """
  <Purpose>
    Generate a digest object given a file object.  The new digest object
    is updated with the contents of 'file_object' prior to returning the
    object to the caller.

  <Arguments>
    file_object:
      File object whose contents will be used as the data
      to update the hash of a digest object to be returned.

    algorithm:
      The hash algorithm (e.g., md5, sha1, sha256).

    hash_library:
      The library providing the hash algorithms
      (e.g., pycrypto, hashlib).

  <Exceptions>
    securesystemslib.exceptions.UnsupportedAlgorithmError

    securesystemslib.exceptions.Error

  <Side Effects>
    Calls securesystemslib.hash.digest() to create the actual digest object.

  <Returns>
    Digest object (e.g., hashlib.new(algorithm) or
    algorithm.new() # pycrypto).
  """

  # Digest object returned whose hash will be updated using 'file_object'.
  # digest() raises:
  # securesystemslib.exceptions.UnsupportedAlgorithmError
  # securesystemslib.exceptions.Error
  digest_object = digest(algorithm, hash_library)

  # Defensively seek to beginning, as there's no case where we don't
  # intend to start from the beginning of the file.
  file_object.seek(0)

  # Read the contents of the file object in at most 4096-byte chunks.
  # Update the hash with the data read from each chunk and return after
  # the entire file is processed.
  while True:
    chunksize = 4096
    data = file_object.read(chunksize)
    if not data:
      break

    if not isinstance(data, six.binary_type):
      digest_object.update(data.encode('utf-8'))

    else:
      digest_object.update(data)

  return digest_object





def digest_filename(filename, algorithm=_DEFAULT_HASH_ALGORITHM,
                    hash_library=_DEFAULT_HASH_LIBRARY):
  """
  <Purpose>
    Generate a digest object, update its hash using a file object
    specified by filename, and then return it to the caller.

  <Arguments>
    filename:
      The filename belonging to the file object to be used.

    algorithm:
      The hash algorithm (e.g., md5, sha1, sha256).

    hash_library:
      The library providing the hash algorithms
      (e.g., pycrypto, hashlib).

  <Exceptions>
    securesystemslib.exceptions.UnsupportedAlgorithmError
    securesystemslib.exceptions.Error

  <Side Effects>
    Calls securesystemslib.hash.digest_fileobject() after opening 'filename'.
    File closed before returning.

  <Returns>
    Digest object (e.g., hashlib.new(algorithm) or
    algorithm.new() # pycrypto).
  """

  # Open 'filename' in read+binary mode.
  file_object = open(filename, 'rb')

  # Create digest_object and update its hash data from file_object.
  # digest_fileobject() raises:
  # securesystemslib.exceptions.UnsupportedAlgorithmError
  # securesystemslib.exceptions.Error
  digest_object = digest_fileobject(file_object, algorithm, hash_library)

  file_object.close()

  return digest_object



#!/usr/bin/env python

"""
<Program Name>
  schema.py

<Author>
  Geremy Condra
  Vladimir Diaz <vladimir.v.diaz@gmail.com>

<Started>
  Refactored April 30, 2012 (previously named checkjson.py). -Vlad

<Copyright>
  See LICENSE for licensing information.

<Purpose>
  Provide a variety of classes that compare objects
  based on their format and determine if they match.
  These classes, or schemas, do not simply check the
  type of the objects being compared, but inspect
  additional aspects of the objects like names and
  the number of items included.
  For example:
  >>> good = {'first': 'Marty', 'last': 'McFly'}
  >>> bad = {'sdfsfd': 'Biff', 'last': 'Tannen'}
  >>> bad = {'sdfsfd': 'Biff', 'last': 'Tannen'}
  >>> schema = Object(first=AnyString(), last=AnyString())
  >>> schema.matches(good)
  True
  >>> schema.matches(bad)
  False
  In the process of determining if the two objects matched the template,
  securesystemslib.schema.Object() inspected the named keys of both dictionaries.
  In the case of the 'bad' dict, a 'first' dict key could not be found.
  As a result, 'bad' was flagged a mismatch.

  'schema.py' provides additional schemas for testing objects based on other
  criteria.  See 'securesystemslib.formats.py' and the rest of this module for extensive
  examples.  Anything related to the checking of TUF objects and their formats
  can be found in 'formats.py'.
"""

# Help with Python 3 compatibility, where the print statement is a function, an
# implicit relative import is invalid, and the '/' operator performs true
# division.  Example:  print 'hello world' raises a 'SyntaxError' exception.
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

import re
import sys

import six

import securesystemslib.exceptions


class Schema:
  """
  <Purpose>
    A schema matches a set of possible Python objects, of types
    that are encodable in JSON.  'Schema' is the base class for
    the other classes defined in this module.  All derived classes
    should implement check_match().
  """

  def matches(self, object):
    """
    <Purpose>
      Return True if 'object' matches this schema, False if it doesn't.
      If the caller wishes to signal an error on a failed match, check_match()
      should be called, which will raise a 'exceptions.FormatError' exception.
    """

    try:
      self.check_match(object)
    except securesystemslib.exceptions.FormatError:
      return False
    else:
      return True


  def check_match(self, object):
    """
    <Purpose>
      Abstract method.  Classes that inherit from 'Schema' must
      implement check_match().  If 'object' matches the schema, check_match()
      should simply return.  If 'object' does not match the schema,
      'exceptions.FormatError' should be raised.
    """

    raise NotImplementedError()





class Any(Schema):
  """
  <Purpose>
    Matches any single object.  Whereas other schemas explicitly state
    the required type of its argument, Any() does not. It simply does a
    'pass' when 'check_match()' is called and at the point where the schema
    is instantiated.
    Supported methods include
      matches(): returns a Boolean result.
      check_match(): passed
  <Example Use>

    >>> schema = Any()
    >>> schema.matches('A String')
    True
    >>> schema.matches([1, 'list'])
    True
  """

  def __init__(self):
    pass


  def check_match(self, object):
    pass





class String(Schema):
  """
  <Purpose>
    Matches a particular string.  The argument object
    must be a string and be equal to a specific string value.
    At instantiation, the string is set and any future comparisons
    are checked against this internal string value.
    Supported methods include
      matches(): returns a Boolean result.
      check_match(): raises 'exceptions.FormatError' on a mismatch.
  <Example Use>

    >>> schema = String('Hi')
    >>> schema.matches('Hi')
    True
    >>> schema.matches('Not hi')
    False
  """

  def __init__(self, string):
    if not isinstance(string, six.string_types):
      raise securesystemslib.exceptions.FormatError('Expected a string but got '+repr(string))

    self._string = string


  def check_match(self, object):
    if self._string != object:
      raise securesystemslib.exceptions.FormatError('Expected '+repr(self._string)+' got '+repr(object))





class AnyString(Schema):
  """
  <Purpose>
    Matches any string, but not a non-string object.  This schema
    can be viewed as the Any() schema applied to Strings, but an
    additional check is performed to ensure only strings are considered.
    Supported methods include
      matches(): returns a Boolean result.
      check_match(): raises 'exceptions.FormatError' on a mismatch.
  <Example Use>

    >>> schema = AnyString()
    >>> schema.matches('')
    True
    >>> schema.matches('a string')
    True
    >>> schema.matches(['a'])
    False
    >>> schema.matches(3)
    False
    >>> schema.matches(u'a unicode string')
    True
    >>> schema.matches({})
    False
  """

  def __init__(self):
    pass


  def check_match(self, object):
    if not isinstance(object, six.string_types):
      raise securesystemslib.exceptions.FormatError('Expected a string but got '+repr(object))





class AnyBytes(Schema):
  """
  <Purpose>
    Matches any byte string, but not a non-byte object.  This schema
    can be viewed as the Any() schema applied to byte strings, but an
    additional check is performed to ensure only strings are considered.
    Supported methods include
      matches(): returns a Boolean result.
      check_match(): raises 'exceptions.FormatError' on a mismatch.
  <Example Use>

    >>> schema = AnyBytes()
    >>> schema.matches(b'')
    True
    >>> schema.matches(b'a string')
    True
    >>> schema.matches(['a'])
    False
    >>> schema.matches(3)
    False
    >>> schema.matches({})
    False
  """

  def __init__(self):
    pass


  def check_match(self, object):
    if not isinstance(object, six.binary_type):
      raise securesystemslib.exceptions.FormatError('Expected a byte string but got '+repr(object))





class LengthString(Schema):
  """
  <Purpose>
    Matches any string of a specified length.  The argument object
    must be a string.  At instantiation, the string length is set
    and any future comparisons are checked against this internal
    string value length.
    Supported methods include
      matches(): returns a Boolean result.
      check_match(): raises 'exceptions.FormatError' on a mismatch.
  <Example Use>

    >>> schema = LengthString(5)
    >>> schema.matches('Hello')
    True
    >>> schema.matches('Hi')
    False
  """

  def __init__(self, length):
    if isinstance(length, bool) or not isinstance(length, six.integer_types):
      # We need to check for bool as a special case, since bool
      # is for historical reasons a subtype of int.
      raise securesystemslib.exceptions.FormatError('Got ' + repr(length) + ' instead of an integer.')

    self._string_length = length


  def check_match(self, object):
    if not isinstance(object, six.string_types):
      raise securesystemslib.exceptions.FormatError('Expected a string but got ' + repr(object))

    if len(object) != self._string_length:
      raise securesystemslib.exceptions.FormatError('Expected a string of length ' + \
                            repr(self._string_length))





class LengthBytes(Schema):
  """
  <Purpose>
    Matches any Bytes of a specified length.  The argument object must be either
    a str() in Python 2, or bytes() in Python 3.  At instantiation, the bytes
    length is set and any future comparisons are checked against this internal
    bytes value length.
    Supported methods include
      matches(): returns a Boolean result.
      check_match(): raises 'exceptions.FormatError' on a mismatch.
  <Example Use>

    >>> schema = LengthBytes(5)
    >>> schema.matches(b'Hello')
    True
    >>> schema.matches(b'Hi')
    False
  """

  def __init__(self, length):
    if isinstance(length, bool) or not isinstance(length, six.integer_types):
      # We need to check for bool as a special case, since bool
      # is for historical reasons a subtype of int.
      raise securesystemslib.exceptions.FormatError('Got ' + repr(length) + ' instead of an integer.')

    self._bytes_length = length


  def check_match(self, object):
    if not isinstance(object, six.binary_type):
      raise securesystemslib.exceptions.FormatError('Expected a byte but got ' + repr(object))

    if len(object) != self._bytes_length:
      raise securesystemslib.exceptions.FormatError('Expected a byte of length ' + \
                            repr(self._bytes_length))





class OneOf(Schema):
  """
  <Purpose>
    Matches an object that matches any one of several schemas.  OneOf()
    returns a result as soon as one of its recognized sub-schemas is encountered
    in the object argument.  When OneOf() is instantiated, its supported
    sub-schemas are specified by a sequence type (e.g., a list, tuple, etc.).
    A mismatch is returned after checking all sub-schemas and not finding
    a supported type.
    Supported methods include
      matches(): returns a Boolean result.
      check_match(): raises 'exceptions.FormatError' on a mismatch.
  <Example Use>
    >>> schema = OneOf([ListOf(Integer()), String('Hello'), String('bye')])
    >>> schema.matches(3)
    False
    >>> schema.matches('bye')
    True
    >>> schema.matches([])
    True
    >>> schema.matches([1,2])
    True
    >>> schema.matches(['Hi'])
    False
  """

  def __init__(self, alternatives):
    # Ensure each item of the list contains the expected object type.
    if not isinstance(alternatives, list):
      raise securesystemslib.exceptions.FormatError('Expected a list but got ' + repr(alternatives))
    for alternative in alternatives:
      if not isinstance(alternative, Schema):
        raise securesystemslib.exceptions.FormatError('List contains an invalid item ' + repr(alternative))

    self._alternatives = alternatives


  def check_match(self, object):
    # Simply return as soon as we find a match.
    # Raise 'exceptions.FormatError' if no matches are found.
    for alternative in self._alternatives:
      if alternative.matches(object):
        return
    raise securesystemslib.exceptions.FormatError('Object did not match a recognized alternative.')





class AllOf(Schema):
  """
  <Purpose>
    Matches the intersection of a list of schemas.  The object being tested
    must match all of the required sub-schemas.  Unlike OneOf(), which can
    return a result as soon as a match is found in one of its supported
    sub-schemas, AllOf() must verify each sub-schema before returning a
    result.
    Supported methods include
      matches(): returns a Boolean result.
      check_match(): raises 'exceptions.FormatError' on a mismatch.
  <Example Use>
    >>> schema = AllOf([Any(), AnyString(), String('a')])
    >>> schema.matches('b')
    False
    >>> schema.matches('a')
    True
  """

  def __init__(self, required_schemas):
    # Ensure each item of the list contains the expected object type.
    if not isinstance(required_schemas, list):
      raise securesystemslib.exceptions.FormatError('Expected a list but got'+repr(required_schemas))
    for schema in required_schemas:
      if not isinstance(schema, Schema):
        raise securesystemslib.exceptions.FormatError('List contains an invalid item '+repr(schema))

    self._required_schemas = required_schemas[:]


  def check_match(self, object):
    for required_schema in self._required_schemas:
      required_schema.check_match(object)





class Boolean(Schema):
  """
  <Purpose>
    Matches a boolean.  The object argument must be one
    of True or False.  All other types are flagged as mismatches.
    Supported methods include
      matches(): returns a Boolean result.
      check_match(): raises 'exceptions.FormatError' on a mismatch.
  <Example Use>
    >>> schema = Boolean()
    >>> schema.matches(True) and schema.matches(False)
    True
    >>> schema.matches(11)
    False
 """

  def __init__(self):
    pass


  def check_match(self, object):
    if not isinstance(object, bool):
      raise securesystemslib.exceptions.FormatError('Got '+repr(object)+' instead of a boolean.')





class ListOf(Schema):
  """
  <Purpose>
    Matches a homogeneous list of some sub-schema.  That is, all the
    sub-schema must be of the same type.  The object argument must
    be a sequence type (e.g., a list, tuple, etc.).  When ListOf()
    is instantiated, a minimum and maximum count can be specified
    for the homogeneous sub-schema list.  If min_count is set to
    'n', the object argument sequence must contain 'n' items.  See
    ListOf()'s __init__ method for the expected arguments.
    Supported methods include
      matches(): returns a Boolean result.
      check_match(): raises 'exceptions.FormatError' on a mismatch.
  <Example Use>
    >>> schema = ListOf(RegularExpression('(?:..)*'))
    >>> schema.matches('hi')
    False
    >>> schema.matches([])
    True
    >>> schema.matches({})
    False
    >>> schema.matches(['Hi', 'this', 'list', 'is', 'full', 'of', 'even', 'strs'])
    True
    >>> schema.matches(['This', 'one', 'is not'])
    False
    >>> schema = ListOf(Integer(), min_count=3, max_count=10)
    >>> schema.matches([3]*2)
    False
    >>> schema.matches([3]*3)
    True
    >>> schema.matches([3]*10)
    True
    >>> schema.matches([3]*11)
    False
  """

  def __init__(self, schema, min_count=0, max_count=sys.maxsize, list_name='list'):
    """
    <Purpose>
      Create a new ListOf schema.
    <Arguments>
      schema:  The pattern to match.
      min_count: The minimum number of sub-schema in 'schema'.
      max_count: The maximum number of sub-schema in 'schema'.
      list_name: A string identifier for the ListOf object.
    """

    if not isinstance(schema, Schema):
      message = 'Expected Schema type but got '+repr(schema)
      raise securesystemslib.exceptions.FormatError(message)

    self._schema = schema
    self._min_count = min_count
    self._max_count = max_count
    self._list_name = list_name


  def check_match(self, object):
    if not isinstance(object, (list, tuple)):
      message = 'Expected '+repr(self._list_name)+' but got '+repr(object)
      raise securesystemslib.exceptions.FormatError(message)

    # Check if all the items in the 'object' list
    # match 'schema'.
    for item in object:
      try:
        self._schema.check_match(item)
      except securesystemslib.exceptions.FormatError as e:
        raise securesystemslib.exceptions.FormatError(str(e)+' in '+repr(self._list_name))

    # Raise exception if the number of items in the list is
    # not within the expected range.
    if not (self._min_count <= len(object) <= self._max_count):
        raise securesystemslib.exceptions.FormatError('Length of '+repr(self._list_name)+' out of range')





class Integer(Schema):
  """
  <Purpose>
    Matches an integer.  A range can be specified.
    For example, only integers between 8 and 42 can be set as
    a requirement.  The object argument is also checked against
    a Boolean type, since booleans have historically been considered
    a sub-type of integer.
    Supported methods include
      matches(): returns a Boolean result.
      check_match(): raises 'exceptions.FormatError' on a mismatch.
  <Example Use>
    >>> schema = Integer()
    >>> schema.matches(99)
    True
    >>> schema.matches(False)
    False
    >>> schema.matches('a string')
    False
    >>> Integer(lo=10, hi=30).matches(25)
    True
    >>> Integer(lo=10, hi=30).matches(5)
    False
  """

  def __init__(self, lo = -2147483648, hi = 2147483647):
    """
    <Purpose>
      Create a new Integer schema.
    <Arguments>
      lo: The minimum value the int object argument can be.
      hi: The maximum value the int object argument can be.
    """

    self._lo = lo
    self._hi = hi


  def check_match(self, object):
    if isinstance(object, bool) or not isinstance(object, six.integer_types):
      # We need to check for bool as a special case, since bool
      # is for historical reasons a subtype of int.
      raise securesystemslib.exceptions.FormatError('Got '+repr(object)+' instead of an integer.')

    elif not (self._lo <= object <= self._hi):
      int_range = '['+repr(self._lo)+', '+repr(self._hi)+'].'
      raise securesystemslib.exceptions.FormatError(repr(object)+' not in range '+int_range)





class DictOf(Schema):
  """
  <Purpose>
    Matches a mapping from items matching a particular key-schema
    to items matching a value-schema (i.e., the object being checked
    must be a dict).  Note that in JSON, keys must be strings.  In the
    example below, the keys of the dict must be one of the letters
    contained in 'aeiou' and the value must be a structure containing
    any two strings.
    Supported methods include
      matches(): returns a Boolean result.
      check_match(): raises 'exceptions.FormatError' on a mismatch.
  <Example Use>
    >>> schema = DictOf(RegularExpression(r'[aeiou]+'), Struct([AnyString(), AnyString()]))
    >>> schema.matches('')
    False
    >>> schema.matches({})
    True
    >>> schema.matches({'a': ['x', 'y'], 'e' : ['', '']})
    True
    >>> schema.matches({'a': ['x', 3], 'e' : ['', '']})
    False
    >>> schema.matches({'a': ['x', 'y'], 'e' : ['', ''], 'd' : ['a', 'b']})
    False
  """

  def __init__(self, key_schema, value_schema):
    """
    <Purpose>
      Create a new DictOf schema.
    <Arguments>
      key_schema:  The dictionary's key.
      value_schema: The dictionary's value.
    """

    if not isinstance(key_schema, Schema):
      raise securesystemslib.exceptions.FormatError('Expected Schema but got '+repr(key_schema))

    if not isinstance(value_schema, Schema):
      raise securesystemslib.exceptions.FormatError('Expected Schema but got '+repr(value_schema))

    self._key_schema = key_schema
    self._value_schema = value_schema


  def check_match(self, object):
    if not isinstance(object, dict):
      raise securesystemslib.exceptions.FormatError('Expected a dict but got '+repr(object))

    for key, value in six.iteritems(object):
      self._key_schema.check_match(key)
      self._value_schema.check_match(value)





class Optional(Schema):
  """
  <Purpose>
    Provide a way for the Object() schema to accept optional
    dictionary keys.  The Object() schema outlines how a dictionary
    should look, such as the names for dict keys and the object type
    of the dict values.  Optional()'s intended use is as a sub-schema
    to Object().  Object() flags an object as a mismatch if a required
    key is not encountered, however, dictionary keys labeled Optional()
    are not required to appear in the object's list of required keys.
    If an Optional() key IS found, Optional()'s sub-schemas are
    then verified.

    Supported methods include
      matches(): returns a Boolean result.
      check_match(): raises 'exceptions.FormatError' on a mismatch.
  <Example Use>
    >>> schema = Object(k1=String('X'), k2=Optional(String('Y')))
    >>> schema.matches({'k1': 'X', 'k2': 'Y'})
    True
    >>> schema.matches({'k1': 'X', 'k2': 'Z'})
    False
    >>> schema.matches({'k1': 'X'})
    True
  """

  def __init__(self, schema):
    if not isinstance(schema, Schema):
      raise securesystemslib.exceptions.FormatError('Expected Schema, but got '+repr(schema))
    self._schema = schema


  def check_match(self, object):
    self._schema.check_match(object)





class Object(Schema):
  """
  <Purpose>
    Matches a dict from specified keys to key-specific types.  Unrecognized
    keys are allowed.  The Object() schema outlines how a dictionary
    should look, such as the names for dict keys and the object type of the
    dict values.  See schema.Optional() to learn how Object() incorporates
    optional sub-schemas.

    Supported methods include
      matches(): returns a Boolean result.
      check_match(): raises 'exceptions.FormatError' on a mismatch.
  <Example Use>
    >>> schema = Object(a=AnyString(), bc=Struct([Integer(), Integer()]))
    >>> schema.matches({'a':'ZYYY', 'bc':[5,9]})
    True
    >>> schema.matches({'a':'ZYYY', 'bc':[5,9], 'xx':5})
    True
    >>> schema.matches({'a':'ZYYY', 'bc':[5,9,3]})
    False
    >>> schema.matches({'a':'ZYYY'})
    False
  """

  def __init__(self, object_name='object', **required):
    """
    <Purpose>
      Create a new Object schema.
    <Arguments>
      object_name: A string identifier for the object argument.

      A variable number of keyword arguments is accepted.
    """

    # Ensure valid arguments.
    for key, schema in six.iteritems(required):
      if not isinstance(schema, Schema):
        raise securesystemslib.exceptions.FormatError('Expected Schema but got '+repr(schema))

    self._object_name = object_name
    self._required = list(required.items())


  def check_match(self, object):
    if not isinstance(object, dict):
      message = 'Wanted a '+repr(self._object_name)+'.'
      raise securesystemslib.exceptions.FormatError(message)

    # (key, schema) = (a, AnyString()) = (a=AnyString())
    for key, schema in self._required:
      # Check if 'object' has all the required dict keys.
      # If not one of the required keys, check if it is an Optional().
      try:
        item = object[key]
      except KeyError:
        # If not an Optional schema, raise an exception.
        if not isinstance(schema, Optional):
          message = 'Missing key ' + repr(key) + ' in ' + repr(self._object_name)
          raise securesystemslib.exceptions.FormatError(message)
      # Check that 'object's schema matches Object()'s schema for this
      # particular 'key'.
      else:
        try:
          schema.check_match(item)
        except securesystemslib.exceptions.FormatError as e:
          raise securesystemslib.exceptions.FormatError(str(e) + ' in ' + self._object_name + '.' + key)





class Struct(Schema):
  """
  <Purpose>
    Matches a non-homogeneous list of items.  The sub-schemas
    are allowed to vary.  The object argument must be a sequence type
    (e.g., a list, tuple, etc.).  There is also an option to specify
    that additional schemas not explicitly defined at instantiation
    are allowed.  See __init__() for the complete list of arguments
    accepted.
    Supported methods include
      matches(): returns a Boolean result.
      check_match(): raises 'exceptions.FormatError' on a mismatch.
  <Example Use>
    >>> schema = Struct([ListOf(AnyString()), AnyString(), String('X')])
    >>> schema.matches(False)
    False
    >>> schema.matches('Foo')
    False
    >>> schema.matches([[], 'Q', 'X'])
    True
    >>> schema.matches([[], 'Q', 'D'])
    False
    >>> schema.matches([[3], 'Q', 'X'])
    False
    >>> schema.matches([[], 'Q', 'X', 'Y'])
    False
    >>> schema = Struct([String('X')], allow_more=True)
    >>> schema.matches([])
    False
    >>> schema.matches(['X'])
    True
    >>> schema.matches(['X', 'Y'])
    True
    >>> schema.matches(['X', ['Y', 'Z']])
    True
    >>> schema.matches([['X']])
    False
    >>> schema = Struct([String('X'), Integer()], [Integer()])
    >>> schema.matches([])
    False
    >>> schema.matches({})
    False
    >>> schema.matches(['X'])
    False
    >>> schema.matches(['X', 3])
    True
    >>> schema.matches(['X', 3, 9])
    True
    >>> schema.matches(['X', 3, 9, 11])
    False
    >>> schema.matches(['X', 3, 'A'])
    False
  """

  def __init__(self, sub_schemas, optional_schemas=[], allow_more=False,
               struct_name='list'):
    """
    <Purpose>
      Create a new Struct schema.
    <Arguments>
      sub_schemas: The sub-schemas recognized.
      optional_schemas: The optional list of schemas.
      allow_more: Specifies that an optional list of types is allowed.
      struct_name: A string identifier for the Struct object.
    """

    # Ensure each item of the list contains the expected object type.
    if not isinstance(sub_schemas, (list, tuple)):
      message = 'Expected Schema but got ' + repr(sub_schemas)
      raise securesystemslib.exceptions.FormatError(message)
    for schema in sub_schemas:
      if not isinstance(schema, Schema):
        raise securesystemslib.exceptions.FormatError('Expected Schema but got ' + repr(schema))

    self._sub_schemas = sub_schemas + optional_schemas
    self._min = len(sub_schemas)
    self._allow_more = allow_more
    self._struct_name = struct_name


  def check_match(self, object):
    if not isinstance(object, (list, tuple)):
      raise securesystemslib.exceptions.FormatError('Expected ' + repr(self._struct_name) + '; got ' + repr(object))
    elif len(object) < self._min:
      raise securesystemslib.exceptions.FormatError('Too few fields in ' + self._struct_name)
    elif len(object) > len(self._sub_schemas) and not self._allow_more:
      raise securesystemslib.exceptions.FormatError('Too many fields in ' + self._struct_name)

    # Iterate through the items of 'object', checking against each schema
    # in the list of schemas allowed (i.e., the sub-schemas and also
    # any optional schemas.  The lenth of 'object' must be less than
    # the length of the required schemas + the optional schemas.  However,
    # 'object' is allowed to be only as large as the length of the required
    # schemas.  In the while loop below, we check against these two cases.
    index = 0
    while index < len(object) and index < len(self._sub_schemas):
      item = object[index]
      schema = self._sub_schemas[index]
      schema.check_match(item)
      index = index + 1





class RegularExpression(Schema):
  """
  <Purpose>
    Matches any string that matches a given regular expression.
    The RE pattern set when RegularExpression is instantiated
    must not be None.  See __init__() for a complete list of
    accepted arguments.
    Supported methods include
      matches(): returns a Boolean result.
      check_match(): raises 'exceptions.FormatError' on a mismatch.
  <Example Use>
    >>> schema = RegularExpression('h.*d')
    >>> schema.matches('hello world')
    True
    >>> schema.matches('Hello World')
    False
    >>> schema.matches('hello world!')
    False
    >>> schema.matches([33, 'Hello'])
    False
  """

  def __init__(self, pattern=None, modifiers=0, re_object=None, re_name=None):
    """
    <Purpose>
      Create a new regular expression schema.
    <Arguments>
      pattern:  The pattern to match, or None if re_object is provided.
      modifiers:  Flags to use when compiling the pattern.
      re_object:  A compiled regular expression object.
      re_name: Identifier for the regular expression object.
    """

    if not isinstance(pattern, six.string_types):
      if pattern is not None:
        raise securesystemslib.exceptions.FormatError(repr(pattern) + ' is not a string.')

    if re_object is None:
      if pattern is None:
        error = 'Cannot compare against an unset regular expression'
        raise securesystemslib.exceptions.FormatError(error)
      if not pattern.endswith('$'):
        pattern += '$'
      re_object = re.compile(pattern, modifiers)
    self._re_object = re_object

    if re_name is None:
      if pattern is not None:
        re_name = 'pattern /' + pattern + '/'
      else:
        re_name = 'pattern'
    self._re_name = re_name


  def check_match(self, object):
    if not isinstance(object, six.string_types) or not self._re_object.match(object):
      raise securesystemslib.exceptions.FormatError(repr(object) + ' did not match ' + repr(self._re_name))



if __name__ == '__main__':
  # The interactive sessions of the documentation strings can
  # be tested by running schema.py as a standalone module.
  # python -B schema.py.
  import doctest
  doctest.testmod()

